#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, ReindexMapError

class Endometrial(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the endometrial dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["2.1", "2.1.1"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "2.1": [
                "acetylproteomics.cct.gz", 
                "clinical.txt", 
                "CNA.cct.gz", 
                "definitions.txt",
                "miRNA.cct.gz", 
                "phosphoproteomics_gene.cct.gz", 
                "phosphoproteomics_site.cct.gz", 
                "proteomics.cct.gz", 
                "somatic_binary.cbt.gz", 
                "somatic.maf.gz", 
                "transcriptomics_circular.cct.gz", 
                "transcriptomics_linear.cct.gz"],
            "2.1.1": [
                "acetylproteomics.cct.gz",
                "clinical.txt",
                "CNA.cct.gz",
                "definitions.txt",
                "miRNA.cct.gz",
                "phosphoproteomics_site.cct.gz",
                "proteomics.cct.gz",
                "somatic_binary.cbt.gz",
                "somatic.maf.gz",
                "transcriptomics_circular.cct.gz",
                "transcriptomics_linear.cct.gz",
                "UCEC_followup_9_12.xlsx"],
        }

        super().__init__(cancer_type="endometrial", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data files into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: 

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file
            df_name = file_name.split(".")[0] # Dataframe name will be the first section of file name; i.e. proteomics.txt.gz becomes proteomics

            # Load the file, based on what it is
            if file_name == "clinical.txt":
                # Fix for reading error on clinical.txt:
                with open(file_path, "r", errors="ignore") as clinical_file:
                    df = pd.read_csv(clinical_file, sep="\t", index_col=0)
                df = df.sort_index()
                self._data[df_name] = df # Maps dataframe name to dataframe

            elif file_name == "definitions.txt":
                with open(file_path, "r") as definitions_file:
                    for line in definitions_file.readlines():
                        line = line.strip()
                        line = line.split("\t")
                        term = line[0]
                        definition = line[1]
                        self._definitions[term] = definition

            elif file_name == "somatic.maf.gz":
                df = pd.read_csv(file_path, sep = "\t")
                split_barcode = df["Tumor_Sample_Barcode"].str.split("_", n=1, expand=True) # The first part of the barcode is the patient id, which we need want to make the index
                df["Tumor_Sample_Barcode"] = split_barcode[0]
                df = df[["Tumor_Sample_Barcode","Hugo_Symbol","Variant_Classification","HGVSp_Short"]]
                df = df.rename({"Tumor_Sample_Barcode":"Patient_ID","Hugo_Symbol":"Gene","Variant_Classification":"Mutation","HGVSp_Short":"Location"}, axis='columns')
                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")
                self._data["somatic_mutation"] = df # Maps dataframe name to dataframe

            elif file_name == "acetylproteomics.cct.gz" or file_name == "phosphoproteomics_site.cct.gz":
                df = pd.read_csv(file_path, sep = "\t", index_col=0)
                df.index = df.index.str.rsplit('-', n=1, expand=True) # Separate the index into a multiindex where the 1st level is the gene, and 2nd is the site
                df.index = df.index.set_names(["Name", "Site"]) # Properly name the levels
                df = df.sort_index()
                df = df.transpose()
                self._data[df_name] = df # Maps dataframe name to dataframe

            elif file_name == 'UCEC_followup_9_12.xlsx' and self._version == "2.1.1":
                df = pd.read_excel(file_path)

                # Replace redundant values for 'not reported' with NaN
                nan_equivalents = ['Not Reported/ Unknown', 'Reported/ Unknown', 'Not Applicable', 'na', 'unknown',
                    'Not Performed', 'Unknown tumor status', 'Unknown', 'Unknown Tumor Status', 'Not specified']
                    
                df = df.replace(nan_equivalents, np.nan)

                # Rename, set, and sort index
                df = df.rename(columns={'Case ID': 'Patient_ID'})
                df = df.set_index("Patient_ID")
                df = df.sort_index()

                self._data["followup"] = df

            else:
                df = pd.read_csv(file_path, sep="\t", index_col=0)
                df = df.transpose()
                df = df.sort_index()
                self._data[df_name] = df # Maps dataframe name to dataframe

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Separate out clinical, derived_molecular, and experimental_design dataframes
        all_clinical = self._data["clinical"]
        clinical = all_clinical[[
            'Proteomics_Participant_ID', 'Case_excluded',  'Proteomics_Tumor_Normal',  'Country',
            'Histologic_Grade_FIGO', 'Myometrial_invasion_Specify', 'Histologic_type', 'Treatment_naive', 'Tumor_purity',
            'Path_Stage_Primary_Tumor-pT', 'Path_Stage_Reg_Lymph_Nodes-pN', 'Clin_Stage_Dist_Mets-cM', 'Path_Stage_Dist_Mets-pM',
            'tumor_Stage-Pathological', 'FIGO_stage', 'LVSI', 'BMI', 'Age', 'Diabetes', 'Race', 'Ethnicity', 'Gender', 'Tumor_Site',
            'Tumor_Site_Other', 'Tumor_Focality', 'Tumor_Size_cm',   'Num_full_term_pregnancies']]
        clinical = clinical.rename(columns={"Proteomics_Participant_ID":"Patient_ID"})
        self._data["clinical"] = clinical

        derived_molecular = all_clinical.drop(['Proteomics_Participant_ID', 'Case_excluded',  'Proteomics_Tumor_Normal',  'Country',
            'Histologic_Grade_FIGO', 'Myometrial_invasion_Specify', 'Histologic_type', 'Treatment_naive', 'Tumor_purity',
            'Path_Stage_Primary_Tumor-pT', 'Path_Stage_Reg_Lymph_Nodes-pN', 'Clin_Stage_Dist_Mets-cM', 'Path_Stage_Dist_Mets-pM',
            'tumor_Stage-Pathological', 'FIGO_stage', 'LVSI', 'BMI', 'Age', 'Diabetes', 'Race', 'Ethnicity', 'Gender', 'Tumor_Site',
            'Tumor_Site_Other', 'Tumor_Focality', 'Tumor_Size_cm',   'Num_full_term_pregnancies',
            'Proteomics_TMT_batch', 'Proteomics_TMT_plex', 'Proteomics_TMT_channel', 'Proteomics_Parent_Sample_IDs',
            'Proteomics_Aliquot_ID', 'Proteomics_OCT', 'WXS_normal_sample_type', 'WXS_normal_filename', 'WXS_normal_UUID', 'WXS_tumor_sample_type', 'WXS_tumor_filename',
            'WXS_tumor_UUID', 'WGS_normal_sample_type', 'WGS_normal_UUID', 'WGS_tumor_sample_type', 'WGS_tumor_UUID', 'RNAseq_R1_sample_type', 'RNAseq_R1_filename', 'RNAseq_R1_UUID',
            'RNAseq_R2_sample_type', 'RNAseq_R2_filename', 'RNAseq_R2_UUID', 'miRNAseq_sample_type', 'miRNAseq_UUID', 'Methylation_available', 'Methylation_quality'], axis=1)
        derived_molecular = derived_molecular.rename(columns={"JAK1_Mutation":"JAK1_Mutation_status"})
        self._data["derived_molecular"] = derived_molecular

        experimental_design = all_clinical[['Proteomics_TMT_batch', 'Proteomics_TMT_plex', 'Proteomics_TMT_channel', 'Proteomics_Parent_Sample_IDs',
            'Proteomics_Aliquot_ID', 'Proteomics_OCT', 'WXS_normal_sample_type', 'WXS_normal_filename', 'WXS_normal_UUID', 'WXS_tumor_sample_type', 'WXS_tumor_filename',
            'WXS_tumor_UUID', 'WGS_normal_sample_type', 'WGS_normal_UUID', 'WGS_tumor_sample_type', 'WGS_tumor_UUID', 'RNAseq_R1_sample_type', 'RNAseq_R1_filename', 'RNAseq_R1_UUID',
            'RNAseq_R2_sample_type', 'RNAseq_R2_filename', 'RNAseq_R2_UUID', 'miRNAseq_sample_type', 'miRNAseq_UUID', 'Methylation_available', 'Methylation_quality']]
        self._data["experimental_design"] = experimental_design

        # Drop all excluded samples from the dataset. They were excluded due to poor sample quality, etc.
        clinical = self._data["clinical"]
        cases_to_drop = clinical[clinical["Case_excluded"] == "Yes"].index.union(clinical[clinical["Case_excluded"] == "Yes"]["Patient_ID"])

        for name in self._data.keys(): # Loop over the keys so we can alter the values without any issues
            df = self._data[name]
            df_filtered = df.drop(index=cases_to_drop, errors="ignore")
            self._data[name] = df_filtered

        # Drop Case_excluded column from clinical, now that we've dropped all excluded cases in the dataset.
        clinical = self._data["clinical"]
        clinical = clinical.drop(columns=["Case_excluded"])

        # Add a Sample_Tumor_Normal column to the clinical dataframe, with just "Tumor" or "Normal" values (unlike the Proteomics_Tumor_Normal column, which gives the different types of normal samples)
        raw_map = clinical["Proteomics_Tumor_Normal"] 
        parsed_map = raw_map.where(raw_map == "Tumor", other="Normal") # Replace various types of normal (Adjacent_normal, Myometrium_normal, etc.) with just "Normal"
        clinical.insert(1, "Sample_Tumor_Normal", parsed_map)

        # Mark the Patient_IDs of the normal samples by appending a ".N" to them
        clinical["Patient_ID"] = clinical["Patient_ID"].where(clinical["Sample_Tumor_Normal"] == "Tumor", other=clinical["Patient_ID"] + ".N")

        # Save our new and improved clinical dataframe!
        self._data["clinical"] = clinical

        # Sort CNV dataframe columns alphabetically
        cna = self._data["CNA"]
        cna_sorted = cna.sort_index(axis=1)
        self._data["CNA"] = cna_sorted

        # Fix dataframe names
        rename_dict = { # Keys are old names, values are new names
            "CNA":"CNV",
            "transcriptomics_linear":"transcriptomics",
            "transcriptomics_circular":"circular_RNA",
            "phosphoproteomics_site":"phosphoproteomics",
            "somatic_binary":"somatic_mutation_binary",
            }
        for old, new in rename_dict.items():
            self._data[new] = self._data[old]
            del self._data[old]

        # Call a function from dataframe_tools.py to reindex all the dataframes with sample IDs instead of patient IDs
        # Skip the followup and somatic_mutation dataframes because they're already indexed with Patient_IDs
        sample_id_to_patient_id_map = self._data["clinical"]["Patient_ID"]
        self._data = reindex_all_sample_id_to_patient_id(self._data, sample_id_to_patient_id_map, skip=["followup", "somatic_mutation"])

        # We no longer need the Patient_ID column in the clinical dataframe, because it's in the index. So we'll remove it.
        clinical = self._data["clinical"]
        clinical = clinical.drop(columns="Patient_ID")
        self._data["clinical"] = clinical

        # Get a union of all dataframes' indices, with duplicates removed
        # Exclude the followup dataframe because it has samples from a different cohort that aren't included anywhere else in the dataset
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset.
        clinical = self._data["clinical"]
        clinical = clinical.reindex(master_index)
        self._data['clinical'] = clinical

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='endometrial carcinoma (uterine)', pmid=32059776)



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime

from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError

class Pdac(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": [
            "clinical_table_140.tsv.gz",
            "microRNA_TPM_log2_Normal.cct.gz",
            "microRNA_TPM_log2_Tumor.cct.gz",
            "meta_table_140.tsv.gz",
            "mRNA_RSEM_UQ_log2_Normal.cct.gz",
            "mRNA_RSEM_UQ_log2_Tumor.cct.gz",
            "PDAC_mutation.maf.gz",
            "phosphoproteomics_site_level_MD_abundance_normal.cct.gz",
            "phosphoproteomics_site_level_MD_abundance_tumor.cct.gz",
            "proteomics_gene_level_MD_abundance_normal.cct.gz",
            "proteomics_gene_level_MD_abundance_tumor.cct.gz",
            "RNA_fusion_unfiltered_normal.tsv.gz",
            "RNA_fusion_unfiltered_tumor.tsv.gz",
            "SCNA_log2_gene_level.cct.gz"],
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="pdac", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below
            mark_normal = lambda s: s + ".N"
            remove_type_tag = lambda s: s[:-2] # remove _T and similar tags from end of string

            if file_name == "clinical_table_140.tsv.gz": # Note that we use the "file_name" variable to identify files. That way we don't have to use the whole path.
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.rename_axis("Patient_ID", axis="index")
                df = df.sort_index()
                df.columns.name = "Name"
                df["Sample_Tumor_Normal"] = "Tumor"
                self._data["clinical"] = df

            elif file_name == "meta_table_140.tsv.gz":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.sort_index()
                df.index.name = "Patient_ID"
                df.columns.name = "Name"
                self._data["derived_molecular"] = df

            elif file_name == "microRNA_TPM_log2_Normal.cct.gz":
                df_normal = pd.read_csv(file_path, sep='\t', index_col=0)
                df_normal = df_normal.sort_index()
                df_normal = df_normal.transpose()
                df_normal = df_normal.rename(index=mark_normal)

                # merge tumor and normal if tumor data has already been read
                if "miRNA" in self._data:
                    df_tumor = self._data["miRNA"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["miRNA"] = df_combined
                else:
                    self._data["miRNA"] = df_normal

            elif file_name == "microRNA_TPM_log2_Tumor.cct.gz":
                df_tumor = pd.read_csv(file_path, sep='\t', index_col=0)
                df_tumor = df_tumor.sort_index()
                df_tumor = df_tumor.transpose()

                # merge tumor and normal if normal data has already been read
                if "miRNA" in self._data:
                    df_normal = self._data["miRNA"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["miRNA"] = df_combined
                else:
                    self._data["miRNA"] = df_tumor

            elif file_name == "mRNA_RSEM_UQ_log2_Normal.cct.gz":
                # create df for normal data
                df_normal = pd.read_csv(file_path, sep='\t', index_col=0)
                df_normal = df_normal.sort_index()
                df_normal = df_normal.transpose()
                df_normal = df_normal.rename(index=mark_normal)
                
                # merge tumor and normal if tumor data has already been read
                if "transcriptomics" in self._data:
                    df_tumor = self._data["transcriptomics"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["transcriptomics"] = df_combined
                else:
                    self._data["transcriptomics"] = df_normal

            elif file_name == "mRNA_RSEM_UQ_log2_Tumor.cct.gz":
                # create df for tumor data
                df_tumor = pd.read_csv(file_path, sep='\t', index_col=0)
                df_tumor = df_tumor.sort_index()
                df_tumor = df_tumor.transpose()

                # merge tumor and normal if normal data has already been read
                if "transcriptomics" in self._data:
                    df_normal = self._data["transcriptomics"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["transcriptomics"] = df_combined
                else:
                    self._data["transcriptomics"] = df_tumor

            elif file_name == "PDAC_mutation.maf.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df[["Hugo_Symbol", "Variant_Classification", "HGVSp_Short", "Tumor_Sample_Barcode"]]
                df = df.rename({"Tumor_Sample_Barcode":"Patient_ID","Hugo_Symbol":"Gene","Variant_Classification":"Mutation","HGVSp_Short":"Location"}, axis='columns')
                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")
                df = df.rename(index=remove_type_tag)
                df.columns.name = "Name"
                self._data["somatic_mutation"] = df

            elif file_name == "phosphoproteomics_site_level_MD_abundance_normal.cct.gz":
                # create df form normal data
                df_normal = pd.read_csv(file_path, sep='\t')
                column_split = df_normal["Index"].str.rsplit("_", n=1, expand=True)
                df_normal = df_normal.assign(
                    Site = column_split[1],
                    Database_ID = column_split[0]
                )
                df_normal = df_normal.drop(columns="Index")
                df_normal = df_normal.rename(columns={"Gene":"Name"})
                df_normal = df_normal.set_index(["Name", "Site", "Peptide", "Database_ID"])
                df_normal = df_normal.sort_index()
                df_normal = df_normal.transpose()
                df_normal = df_normal.rename(index=mark_normal)

                # merge tumor and normal if tumor data has already been read
                if "phosphoproteomics" in self._data:
                    df_tumor = self._data["phosphoproteomics"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    #df_combined.columns.name = "Name"
                    self._data["phosphoproteomics"] = df_combined
                else:
                    self._data["phosphoproteomics"] = df_normal

            elif file_name == "phosphoproteomics_site_level_MD_abundance_tumor.cct.gz":
                df_tumor = pd.read_csv(file_path, sep='\t')
                column_split = df_tumor["Index"].str.rsplit("_", n=1, expand=True)
                df_tumor = df_tumor.assign(
                    Site = column_split[1],
                    Database_ID = column_split[0]
                )
                df_tumor = df_tumor.drop(columns="Index")
                df_tumor = df_tumor.rename(columns={"Gene":"Name"})
                df_tumor = df_tumor.set_index(["Name", "Site", "Peptide", "Database_ID"])
                df_tumor = df_tumor.sort_index()
                df_tumor = df_tumor.transpose()
                
                # merge tumor and normal if normal data has already been read
                if "phosphoproteomics" in self._data:
                    df_normal = self._data["phosphoproteomics"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    #df_combined.columns.name = "Name"
                    self._data["phosphoproteomics"] = df_combined
                else:
                    self._data["phosphoproteomics"] = df_tumor
        
            elif file_name == "proteomics_gene_level_MD_abundance_normal.cct.gz":
                df_normal = pd.read_csv(file_path, sep='\t', index_col=0)
                df_normal = df_normal.sort_index()
                df_normal = df_normal.transpose()
                df_normal = df_normal.rename(index=mark_normal)

                # merge tumor and normal if tumor data has already been read
                if "proteomics" in self._data:
                    df_tumor = self._data["proteomics"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["proteomics"] = df_combined
                else:
                    self._data["proteomics"] = df_normal

            elif file_name == "proteomics_gene_level_MD_abundance_tumor.cct.gz":
                df_tumor = pd.read_csv(file_path, sep='\t', index_col=0)
                df_tumor = df_tumor.sort_index()
                df_tumor = df_tumor.transpose()

                # merge tumor and normal if normal data has already been read
                if "proteomics" in self._data:
                    df_normal = self._data["proteomics"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["proteomics"] = df_combined
                else:
                    self._data["proteomics"] = df_tumor

            elif file_name == "RNA_fusion_unfiltered_normal.tsv.gz":
                df_normal = pd.read_csv(file_path, sep='\t', index_col=0)
                df_normal = df_normal.rename(columns={"Sample": "Patient_ID"})
                df_normal = df_normal.set_index("Patient_ID")
                df_normal = df_normal.rename(index=mark_normal)

                if "gene_fusion" in self._data:
                    df_tumor = self._data ["gene_fusion"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["gene_fusion"] = df_combined
                else:
                    self._data["gene_fusion"] = df_normal

            elif file_name == "RNA_fusion_unfiltered_tumor.tsv.gz":
                df_tumor = pd.read_csv(file_path, sep='\t', index_col=0)
                df_tumor = df_tumor.rename(columns={"Sample": "Patient_ID"})
                df_tumor = df_tumor.set_index("Patient_ID")

                if "gene_fusion" in self._data:
                    df_normal = self._data ["gene_fusion"]
                    df_combined = pd.concat([df_normal, df_tumor])
                    df_combined.index.name = "Patient_ID"
                    df_combined.columns.name = "Name"
                    self._data["gene_fusion"] = df_combined
                else:
                    self._data["gene_fusion"] = df_tumor

            elif file_name == "SCNA_log2_gene_level.cct.gz":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.transpose()
                df.columns.name = "Name"
                df.index.name = "Patient_ID"
                self._data["CNV"] = df


        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data) 

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        new_clinical = self._data["clinical"]
        new_clinical = new_clinical.reindex(master_index)

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = new_clinical

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

        # Print password access only warning
        warnings.warn("The pdac data is currently strictly reserved for CPTAC investigators. "
            "Otherwise, you are not authorized to access these data. "
            "Additionally, even after these data become publicly available, "
            "they will be subject to a publication embargo "
            "(see https://proteomics.cancer.gov/data-portal/about/data-use-agreement "
            "or enter cptac.embargo() to open the webpage for more details).", 
            PublicationEmbargoWarning, stacklevel=2)
    def how_to_cite(self):
        return super().how_to_cite(cancer_type='pancreatic ductal adenocarcinoma', pmid='', unpublished=True)


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .pancandataset import PancanDataset

from .mssmclinical import MssmClinical
from .pdclscc import PdcLscc
from .umichlscc import UmichLscc
from .washulscc import WashuLscc
from .bcmlscc import BcmLscc
from .broadlscc import BroadLscc
from .harmonized import Harmonized
from .joiningdataset import JoiningDataset

#List sources to be downloaded
SOURCES = [
    "bcmlscc",
    "mssmclinical",
    "pdclscc",
    "umichlscc",
    "washulscc",
    "broadlscc",
    "harmonized"
]

class PancanLscc(PancanDataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all the data sources with LSCC data and provide an interface to them."""

        super().__init__(cancer_type="pancanlscc", version=version, no_internet=no_internet)
        
        self._datasets["bcm"] = BcmLscc(no_internet=no_internet, version=self._get_version("bcm"))
        self._datasets["broad"] = BroadLscc(no_internet=no_internet, version=self._get_version("broad"))
        self._datasets["mssm"] = MssmClinical(no_internet=no_internet, version=self._get_version("mssm"), filter_type='pancanlscc')
        self._datasets["pdc"] = PdcLscc(no_internet=no_internet, version=self._get_version("pdc"))
        self._datasets["umich"] = UmichLscc(no_internet=no_internet, version=self._get_version("umich"))
        self._datasets["washu"] = WashuLscc(no_internet=no_internet, version=self._get_version("washu"))
        self._datasets["harmonized"] = Harmonized(no_internet=no_internet, version=self._get_version("harmonized"), filter_type='pancanlscc')
        
        join_dict = {k: v._data for k, v in self._datasets.items()}
        self._joining_dataset = JoiningDataset(join_dict)
        
        self._pancan_unionize_indices()



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import webbrowser
import os.path as path
import io
import sys
import warnings

# Function imports
from .file_download import download
from .file_download import download_text as _download_text
from .exceptions import CptacError, CptacWarning, InvalidParameterError, NoInternetError, OldPackageVersionWarning

# Dataset imports
from .brca import Brca
from .ccrcc import Ccrcc
from .colon import Colon
from .endometrial import Endometrial
from .gbm import Gbm
from .hnscc import Hnscc
from .lscc import Lscc
from .luad import Luad
from .ovarian import Ovarian
from .pdac import Pdac
from .ucecconf import UcecConf

def list_datasets():
    """List all available datasets."""

    dataset_list_url = "https://byu.box.com/shared/static/5vwsvgu8fyzao0pb7rx8lt26huofcdax.tsv"

    try:
        dataset_list_text = _download_text(dataset_list_url)
    except NoInternetError:
        raise NoInternetError("Insufficient internet to download available dataset info. Check your internet connection.") from None

    return pd.read_csv(io.StringIO(dataset_list_text), sep="\t", index_col=0)

def embargo():
    """Open CPTAC embargo details in web browser."""
    message = "Opening embargo details in web browser..."
    print(message, end = '\r')
    webbrowser.open("https://proteomics.cancer.gov/data-portal/about/data-use-agreement")
    print(" " * len(message), end='\r') # Erase the message

def version():
    """Return version number of cptac package."""
    version = {}
    path_here = path.abspath(path.dirname(__file__))
    version_path = path.join(path_here, "version.py")
    with open(version_path) as fp:
        exec(fp.read(), version)
    return(version['__version__'])

def how_to_cite():
    """Give instructions for citing CPTAC datasets."""
    print("If you use the API to generate results, please cite our manuscript describing the API - Lindgren et al. 2021, PMID:33560848, https://pubs.acs.org/doi/10.1021/acs.jproteome.0c00919")
    print('\n')
    print("For instructions on how to cite a specific dataset, please call its how_to_cite method, e.g. cptac.Endometrial().how_to_cite()")

# Helper functions for handling exceptions and warnings
def _exception_handler(exception_type, exception, traceback, default_hook=sys.excepthook): # Because Python binds default arguments when the function is defined, default_hook's default will always refer to the original sys.excepthook
    """Catch cptac-generated exceptions, and make them prettier."""
    if issubclass(type(exception), CptacError):
        print(f"cptac error: {str(exception)} ({traceback.tb_frame.f_code.co_filename}, line {traceback.tb_lineno})", file=sys.stderr) # We still send to stderr
    else:
        default_hook(exception_type, exception, traceback) # This way, exceptions from other packages will still be treated the same way

def _warning_displayer(message, category, filename, lineno, file=None, line=None, default_displayer=warnings.showwarning): # Python binds default arguments when the function is defined, so default_displayer's default will always refer to the original warnings.showwarning
    """Catch cptac-generated warnings and make them prettier."""
    if issubclass(category, CptacWarning):
        print(f"cptac warning: {str(message)} ({filename}, line {lineno})", file=sys.stderr) # We still send to stderr
    else:
        default_displayer(message, category, filename, lineno, file, line) # This way, warnings from other packages will still be displayed the same way

sys.excepthook = _exception_handler # Set our custom exception hook
warnings.showwarning = _warning_displayer # And our custom warning displayer
warnings.simplefilter("always", category=CptacWarning) # Edit the warnings filter to show multiple occurences of cptac-generated warnings

# Check whether the package is up-to-date
version_url = "https://byu.box.com/shared/static/kbwivmqnrdnn5im2gu6khoybk5a3rfl0.txt"
try:
    remote_version = _download_text(version_url)
except NoInternetError:
    pass
else:
    local_version = version()
    if remote_version != local_version:
        warnings.warn(f"Your version of cptac ({local_version}) is out-of-date. Latest is {remote_version}. Please run 'pip install --upgrade cptac' to update it.", OldPackageVersionWarning, stacklevel=2)



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import cptac
import cptac.exceptions as ex
import pandas as pd
import logging
import cptac.utils as ut
from ..dataframe_tools import unionize_indices, generate_sample_status_col, sort_all_rows_pancan


class PancanDataset:

    def __init__(self, cancer_type, version, no_internet):
        """Initialize variables for a PancanDataset object."""

        self._cancer_type = cancer_type
        self._version = version
        self._datasets = {} # Child class __init__ needs to fill this
        self._joining_dataset = None
        
        #ignore logging messages
        logger = logging.getLogger()
        logger.setLevel(logging.CRITICAL)


    # Clinical table getters
    def get_clinical(self, source='mssm', tissue_type="both", imputed=False):
        """Get the clinical dataframe from the specified data source."""
        return self._get_dataframe("clinical", source, tissue_type, imputed=imputed)

    def get_demographic(self, source='mssm', tissue_type="both", imputed=False):
        """Get the demographic dataframe from the specified data source."""
        return self._get_dataframe("demographic", source, tissue_type, imputed=imputed)

    def get_medical_conditions(self, source='mssm', tissue_type="both", imputed=False):
        """Get the medical_conditions dataframe from the specified data source."""
        return self._get_dataframe("medical_conditions", source, tissue_type, imputed=imputed)

    def get_previous_cancer(self, source='mssm', tissue_type="both", imputed=False):
        """Get the previous_cancer dataframe from the specified data source."""
        return self._get_dataframe("previous_cancer", source, tissue_type, imputed=imputed)

    def get_cancer_diagnosis(self, source='mssm', tissue_type="both", imputed=False):
        """Get the cancer_diagnosis dataframe from the specified data source."""
        return self._get_dataframe("cancer_diagnosis", source, tissue_type, imputed=imputed)

    def get_followup(self, source='mssm', tissue_type="both", imputed=False):
        """Get the followup dataframe from the specified data source."""
        return self._get_dataframe("followup", source, tissue_type, imputed=imputed)

    # Quantitative table getters
    def get_acetylproteomics(self, source, tissue_type="both", imputed=False):
        """Get the acetylproteomics dataframe from the specified data source."""
        return self._get_dataframe("acetylproteomics", source, tissue_type, imputed=imputed)

    def get_ubiquitylomics(self, source, tissue_type="both", imputed=False):
        """Get the ubiquitylomics dataframe from the specified data source."""
        return self._get_dataframe("ubiquitylomics", source, tissue_type, imputed=imputed)

    def get_circular_RNA(self, source="bcm", tissue_type="both", imputed=False):
        """Get a circular RNA dataframe from the specified data source."""
        return self._get_dataframe("circular_RNA", source, tissue_type, imputed=imputed)

    def get_CNV(self, source="washu", tissue_type="both", imputed=False):
        """Get a CNV dataframe from the specified data source."""
        return self._get_dataframe("CNV", source, tissue_type, imputed=imputed)

    def get_deconvolution(self, source='washu', deconv_algorithm=None, tissue_type="both", imputed=False):
        """Get a deconvolution dataframe from the specified data source.
        
        Parameters:
        deconv_algorithm (str):  Choose an alorithm. Acceptable values are ['cibersort', 'xcell']. 
        source (str): Select data generated by a certain institution. Available sources are ['washu']. Defaults to 'washu'. 
        """
        valid_algs = ['cibersort', 'xcell']
        if deconv_algorithm is None or deconv_algorithm not in valid_algs:
            raise ex.InvalidParameterError(f"Please pass a valid value to the 'deconv_algorithm' parameter to specify which algorithm you want deconvolution data from. Valid options are {valid_algs}.")

        return self._get_dataframe(deconv_algorithm, source, tissue_type, imputed=imputed)

    def get_miRNA(self, source='washu', miRNA_type='total', tissue_type="both", imputed=False):
        """Get miRNA dataframe from the specified data source.
        
        Parameters:
        source (str): Select data generated by a certain institution. Available sources are ['washu']. Defaults to 'washu'.
        miRNA_type (str): Choose the type of miRNA molecules measured. Acceptable values are ['total', 'precursor', 'mature'].
        """
        return self._get_dataframe(miRNA_type+'_miRNA', source, tissue_type, imputed=imputed)
    
    def get_phosphoproteomics(self, source="umich", tissue_type="both", imputed=False):
        """Get the phosphoproteomics dataframe from the specified data source."""
        return self._get_dataframe("phosphoproteomics", source, tissue_type, imputed=imputed)

    def get_proteomics(self, source="umich", tissue_type="both", imputed=False):
        """Get the proteomics dataframe from the specified data source."""
        return self._get_dataframe("proteomics", source, tissue_type, imputed=imputed)

    def get_somatic_mutation(self, source="harmonized", tissue_type="both", imputed=False):
        """Get the somatic mutation dataframe from the specified data source."""
        ''' source (str): Select data generated by a certain institution. Default is harmonized. "washu" is also available'''
        return self._get_dataframe("somatic_mutation", source, tissue_type, imputed=imputed)
    
    def get_transcriptomics(self, source, tissue_type="both", imputed=False):
        """Get the transcriptomics dataframe from the specified data source."""
        """source (str): Select data generated by a certain institution. Available sources are ['washu','bcm','broad']."""
        return self._get_dataframe("transcriptomics", source, tissue_type, imputed=imputed)
    
    def get_tumor_purity(self, source='washu', tissue_type="both", imputed=False):
        """Get the tumor purity dataframe from the specified data source."""
        return self._get_dataframe("tumor_purity", source, tissue_type, imputed=imputed)
    
    def get_docs(self, data, source='washu', tissue_type="both", imputed=False):
        """Get the readme docs for the specified data type and source."""
        if source in self._datasets.keys():
            obj = self._datasets[source] 
            print('README file for ' + data, '\n')
            print(obj._readme_files['readme_'+data])
            return 0
        else:
            raise ex.DataSourceNotFoundError(f"Data source {source} not found for the {self._cancer_type} dataset.")

    # Join functions
    def join_omics_to_omics(
        self, 
        df1_name, 
        df2_name,
        df1_source,
        df2_source,
        genes1=None, 
        genes2=None, 
        how="outer", 
        quiet=False, 
        tissue_type="both"
    ):

        df1_name = f"{df1_source}_{df1_name}"
        df2_name = f"{df2_source}_{df2_name}"
        
        return self._joining_dataset.join_omics_to_omics(
            df1_name=df1_name, 
            df2_name=df2_name,
            genes1=genes1, 
            genes2=genes2, 
            how=how, 
            quiet=quiet, 
            tissue_type=tissue_type
        )
    
    def join_omics_to_mutations(
        self,
        omics_df_name,
        omics_source,
        mutations_genes,  
        omics_genes=None, 
        mutations_filter=None, 
        show_location=True, 
        how="outer", 
        quiet=False,
        tissue_type="both",
        mutation_cols=["Mutation","Location"]
     
    ):
        omics_df_name = f"{omics_source}_{omics_df_name}"
      
        return self._joining_dataset.join_omics_to_mutations(
            omics_df_name = omics_df_name,
            mutations_genes = mutations_genes,
            omics_genes=omics_genes,
            mutations_filter=mutations_filter,
            show_location=show_location,
            how=how,
            quiet=quiet,
            tissue_type=tissue_type,
            mutation_cols=mutation_cols
        )
         
        
    # Help functions
    def get_cancer_type(self):
        return self._cancer_type
    
    def list_data_sources(self):
        """Print which sources provide each data type.
        
        Parameters:
        print_list (bool, optional): Whether to print the list. Default is True. Otherwise, it's returned as a string.
        """

        # This dict will be keyed by data type, and the values will be each source that provides that data type
        data_sources = {}

        for source in sorted(self._datasets.keys()):
            for df_name in sorted(self._datasets[source]._data.keys()):

                if df_name in ["cibersort", "xcell"]:
                    df_name = f"deconvolution_{df_name}" # For clarity

                if df_name in data_sources.keys():
                    data_sources[df_name][0] += f", {source}"
                else:
                    data_sources[df_name] = [source]

        data_sources = pd.\
        DataFrame(data_sources).\
        transpose().\
        sort_index().\
        reset_index()

        data_sources.columns=["Data type", "Available sources"]

        return data_sources
        


    # "Private" methods
    def _get_dataframe(self, name, source, tissue_type, imputed):
        """Check that a given dataframe from a given source exists, and return a copy if it does."""

        if imputed:
            name = name + "_imputed"

        if source in self._datasets.keys():
            return self._datasets[source]._get_dataframe(name, tissue_type)
        else:
            raise ex.DataSourceNotFoundError(f"Data source {source} not found for the {self._cancer_type} dataset.")

    def _get_version(self, source):
        if self._version == "latest":
            return self._version
        elif type(self._version) is dict:
            return self._version[source]
        else:
            return self._version
        
    def _pancan_unionize_indices(self):
        '''Gets a master index of all IDs, adds IDs to clinical (clinical will have all IDs in other data), 
        adds Sample_Tumor_Normal column, and sorts clinical rows'''
        master_index = pd.Index([])
        
        for name, ds in self._datasets.items():
            ds_index = unionize_indices(ds._data) # unionize_indices takes 
            
            master_index = master_index.\
            union(ds_index).\
            drop_duplicates()
   
        # For some cases, there needs to be a multiindex with patient and aliquot IDs. In this case, we drop
        # tuples so they are not included in the master index
        master_index = [x for x in master_index if not isinstance(x, tuple)] # drop tuples from index (Patient_ID, Aliquot)
            
        new_clinical = self._datasets["mssm"]._data["clinical"] # Get the clinical dataframe
        new_clinical = new_clinical.reindex(master_index) # Take the big, comprehensive index generated by unionize_indices and use it to reindex the clinical dataframe. This will insert empty rows for any samples in the master index that weren't in the clinical dataframe before.
        sample_status_col = generate_sample_status_col(new_clinical, normal_test=lambda sample: sample[-2:] == '.N') # Parse the patient IDs and based on that generate a column that says which rows are tumor or normal. Note that if the normal IDs are marked by some format other than a ".N" at the end, you'll need to edit the "normal_test" parameter. But otherwise you're fine.
        new_clinical.insert(0, "Sample_Tumor_Normal", sample_status_col) # Add this new column into the clinical dataframe
        self._datasets["mssm"]._data['clinical'] = new_clinical # Save the edited clinical dataframe
        self._datasets["mssm"]._data = sort_all_rows_pancan(self._datasets["mssm"]._data) # sort clinical after adding master

        
    def get_genotype_all_vars(self, mutations_genes, omics_source, mutations_filter=None, show_location=True, mutation_hotspot=None):
        """Return a dataframe that has the mutation type and wheather or not it is a multiple mutation
        Parameters:
        mutation_genes (str, or list or array-like of str): The gene(s) to get mutation data for.
        mutations_filter (list, optional):  List of mutations to prioritize when filtering out multiple mutations, in order of priority.
        omics_source(str): Source of omics data ex "bcm", "washu", "broad", "umich"
        show_location (bool, optional): Whether to include the Location column from the mutation dataframe. Defaults to True.
        mutation_hotspot (optional): a list of hotspots
        """

        #If they don't give us a filter, this is the default.
           
        mutations_filter = ["Deletion",
                                    'Frame_Shift_Del', 'Frame_Shift_Ins', 'Nonsense_Mutation', 'Nonstop_Mutation', #tuncation
                                    'Missense_Mutation_hotspot',
                                    'Missense_Mutation',
                                    'Amplification',
                                    'In_Frame_Del', 'In_Frame_Ins', 'Splice_Site' ,
                                    'De_Novo_Start_Out_Frame' ,'De_Novo_Start_In_Frame', 
                                    'Start_Codon_Ins', 'Start_Codon_SNP', 
                                    'Silent',
                                    'Wildtype']

        truncations = ['Frame_Shift_Del', 'Frame_Shift_Ins', 'Nonsense_Mutation', 'Nonstop_Mutation', 'Splice_Site']
        missenses = ['In_Frame_Del', 'In_Frame_Ins', 'Missense_Mutation']
        noncodings = ["Intron", "RNA", "3'Flank", "Splice_Region", "5'UTR", "5'Flank", "3'UTR"]



        #check that gene is in the somatic_mutation DataFrame
        somatic_mutation = self.get_somatic_mutation()
        if mutations_genes not in somatic_mutation["Gene"].unique(): #if the gene isn't in the somacic mutations df it will still have CNV data that we want
            def add_del_and_amp_no_somatic(row):
                if row[mutations_genes] <= -.2:
                    mutations = 'Deletion'

                elif row[mutations_genes] >= .2:
                    mutations = 'Amplification'
                else:
                    mutations = "No_Mutation"

                return mutations
            
            cnv = self.get_CNV(source = omics_source)
            #drop the database index from ccrcc and brca
            if isinstance(cnv.keys(), pd.core.indexes.multi.MultiIndex):
                drop = ['Database_ID']
                cnv = ut.reduce_multiindex(df=cnv, levels_to_drop=drop)       
            gene_cnv = cnv[[mutations_genes]]
            mutation_col = gene_cnv.apply(add_del_and_amp_no_somatic, axis=1)
            df = gene_cnv.assign(Mutation = mutation_col)
            return df


        #combine the cnv and mutations dataframe
        combined = self.join_omics_to_mutations(omics_df_name="CNV", mutations_genes=mutations_genes, omics_genes=mutations_genes, omics_source = omics_source)
                

        #drop the database index 
        drop = ['Database_ID']
        combined = ut.reduce_multiindex(df=combined, levels_to_drop=drop)


        #If there are hotspot mutations, append 'hotspot' to the mutation type so that it's prioritized correctly
        def mark_hotspot_locations(row):
            #iterate through each location in the current row
            mutations = []
            for location in row[mutations_genes+'_Location']:
                if location in mutation_hotspot: #if it's a hotspot mutation
                    #get the position of the location
                    position = row[mutations_genes+'_Location'].index(location)
                    #use that to change the correct mutation
                    mutations.append(row[mutations_genes+"_Mutation"][position] + "_hotspot")
                else:
                    # get the position of the location
                    position = row[mutations_genes+'_Location'].index(location)
                    mutations.append(row[mutations_genes+"_Mutation"][position])
            return mutations

        if mutation_hotspot is not None:
            combined['hotspot'] = combined.apply(mark_hotspot_locations, axis=1)
            combined[mutations_genes+"_Mutation"] = combined['hotspot']
            combined = combined.drop(columns='hotspot')
     

        # Based on cnv make a new column with mutation type that includes deletions and amplifications
        def add_del_and_amp(row):
            if row[mutations_genes+ "_" + omics_source + "_CNV"] <= -.2:
                mutations = row[mutations_genes+"_Mutation"] + ['Deletion']
                locations = row[mutations_genes+'_Location']+['Deletion']

            elif row[mutations_genes + "_" + omics_source+"_CNV"] >= .2:
                mutations = row[mutations_genes+"_Mutation"] + ['Amplification']
                locations = row[mutations_genes+'_Location']+['Amplification']
            else:
                mutations = row[mutations_genes+"_Mutation"]
                locations = row[mutations_genes+"_Location"]

            return mutations, locations


        combined['mutations'], combined['locations'] = zip(*combined.apply(add_del_and_amp, axis=1))
       
        #now that we have the deletion and amplifications, we need to prioritize the correct mutations.
        def sort(row):
            sortedcol = []
            location = []
            chosen_indices = []
            sample_mutations_list = row['mutations']
            sample_locations_list = row['locations']
            if len(sample_mutations_list) == 1: #if there's only one mutation in the list
                sortedcol.append(sample_mutations_list[0])
                location.append(sample_locations_list[0])

            else:
                for filter_val in mutations_filter: # This will start at the beginning of the filter list, thus filters earlier in the list are prioritized, like we want
                    if filter_val in sample_mutations_list:
                        chosen_indices = [index for index, value in enumerate(sample_mutations_list) if value == filter_val]
                    if len(chosen_indices) > 0: # We found at least one mutation from the filter to prioritize, so we don't need to worry about later values in the filter priority list
                        break

                if len(chosen_indices) == 0: # None of the mutations for the sample were in the filter, so we're going to have to use our default hierarchy
                    for mutation in sample_mutations_list:
                        if mutation in truncations:
                            chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

                if len(chosen_indices) == 0: # None of them were in the filter, nor were truncations, so we'll grab all the missenses
                    for mutation in sample_mutations_list:
                        if mutation in missenses:
                            chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

                if len(chosen_indices) == 0: # None of them were in the filter, nor were truncations, nor missenses, so we'll grab all the noncodings
                    for mutation in sample_mutations_list:
                        if mutation in noncodings:
                            chosen_indices += [index for index, value in enumerate(sample_mutations_list) if value == mutation]

                soonest_mutation = sample_mutations_list[chosen_indices[0]]
                soonest_location = sample_locations_list[chosen_indices[0]]
                chosen_indices.clear()
                sortedcol.append(soonest_mutation)
                location.append(soonest_location)

            return pd.Series([sortedcol, location],index=['mutations', 'locations'])

        df = combined.apply(sort, axis=1)
        combined['Mutation'] = df['mutations']
        combined['Location'] = df['locations']

        #get a sample_status column that says if the gene has multiple mutations (including dletion and amplification)
        def sample_status(row):
            if len(row['mutations']) > 1: #if there's more than one mutation
                if len(row['mutations']) == 2 and "Wildtype_Tumor" in row['mutations']: #one of the mutations might be a "wildtype tumor"
                    status ="Single_mutation"

                elif len(row['mutations']) == 2 and "Wildtype_Normal" in row['mutations']:
                    status ="Single_mutation"

                else:
                    status = "Multiple_mutation"
            else:
                if row["mutations"] == ["Wildtype_Normal"]:
                    status = "Wildtype_Normal"
                elif row['mutations'] == ['Wildtype_Tumor']:
                    status = "Wildtype_Tumor"
                else:
                    status = "Single_mutation"

            return status
        combined['Mutation_Status'] = combined.apply(sample_status, axis=1)

        #drop all the unnecessary Columns
        df = combined.drop(columns=[mutations_genes+ "_" + omics_source +"_CNV", mutations_genes+"_Mutation", mutations_genes+"_Location", mutations_genes+"_Mutation_Status", 'Sample_Status', 'mutations','locations'])
        df['Mutation'] = [','.join(map(str, l)) for l in df['Mutation']]
        df['Location'] = [','.join(map(str, l)) for l in df['Location']]
        if show_location == False: df = df.drop(columns="Location") #if they don't want us to show the location, drop it
       
        return df



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, ReindexMapError

class Brca(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the brca dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["3.1", "3.1.1", "5.4"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "3.1": [
                "prosp-brca-v3.1-acetylome-ratio-norm-NArm.gct.gz",
                "prosp-brca-v3.1-gene-level-cnv-gistic2-all_data_by_genes.gct.gz",
                "prosp-brca-v3.1-phosphoproteome-ratio-norm-NArm.gct.gz",
                "prosp-brca-v3.1-proteome-ratio-norm-NArm.gct.gz",
                "prosp-brca-v3.1-rnaseq-fpkm-log2-row-norm-2comp.gct.gz",
                "prosp-brca-v3.1-sample-annotation.csv.gz"],
            "3.1.1": [
                "Breast_One_Year_Clinical_Data_20160927.xls",
                "prosp-brca-v3.0-v1.4.somatic.variants.070918.maf.gz",
                "prosp-brca-v3.1-acetylome-ratio-norm-NArm.gct.gz",
                "prosp-brca-v3.1-gene-level-cnv-gistic2-all_data_by_genes.gct.gz",
                "prosp-brca-v3.1-phosphoproteome-ratio-norm-NArm.gct.gz",
                "prosp-brca-v3.1-proteome-ratio-norm-NArm.gct.gz",
                "prosp-brca-v3.1-rnaseq-fpkm-log2-row-norm-2comp.gct.gz",
                "prosp-brca-v3.1-sample-annotation.csv.gz"],
            "5.4" : [
                "prosp-brca-v3.0-v1.4.somatic.variants.070918.maf.gz",
                "prosp-brca-v5.4-public-acetylome-ratio-norm-NArm.gct.gz", 
                "prosp-brca-v5.4-public-gene-level-cnv-gistic2-all_data_by_genes.gct.gz", 
                #"prosp-brca-v5.4-public-immune-profiling-scores-combined.gct.gz", 
                "prosp-brca-v5.4-public-phosphoproteome-ratio-norm-NArm.gct.gz", 
                "prosp-brca-v5.4-public-proteome-ratio-norm-NArm.gct.gz", 
                #"prosp-brca-v5.4-public-rnaseq-fpkm-log2-row-norm-median-mad.gct.gz", 
                "prosp-brca-v5.4-public-rnaseq-fpkm-log2.gct.gz", 
                "prosp-brca-v5.4-public-sample-annotation.csv.gz"]
        }

        super().__init__(cancer_type="brca", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file

            if file_name in ["prosp-brca-v3.1-acetylome-ratio-norm-NArm.gct.gz", 
                            "prosp-brca-v5.4-public-acetylome-ratio-norm-NArm.gct.gz"]:
                df = pd.read_csv(file_path, sep='\t', skiprows=2, dtype=object) # First two rows of file aren't part of the dataframe. Also, due to extra metadata rows we're going to remove, all cols have mixed types, so we pass dtype=object for now.
                df = df[df["GeneSymbol"] != "na"] # There are several metadata rows at the beginning of the dataframe, which duplicate the clinical and derived_molecular dataframes. They all don't have a value for GeneSymbol, so we'll use that to filter them out.

                # Prepare some columns we'll need later for the multiindex
                df["variableSites"] = df["variableSites"].str.replace(r"[a-z\s]", "", regex=True) # Get rid of all lowercase delimeters and whitespace in the sites
                df = df.rename(columns={
                    "GeneSymbol": "Name",
                    "variableSites": "Site",
                    "sequence": "Peptide", # We take this instead of sequenceVML, to match the other datasets' format
                    "accession_numbers": "Database_ID" # We take all accession numbers they have, instead of the singular accession_number column
                    })

                # Some rows have at least one localized acetylation site, but also have other acetylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                split_ids = df["id"].str.split('_', expand=True)
                unlocalized_to_drop = df.index[~split_ids[3].eq(split_ids[4]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 3 of the split "id" column is number of phosphorylations detected, and column 4 is number of phosphorylations localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])                

                df = df.drop(columns=["id", "id.description", "geneSymbol", "numColumnsVMsiteObserved", "bestScore", "bestDeltaForwardReverseScore", 
                "Best_scoreVML", "sequenceVML", "accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA",
                "protein_mw", "species", "speciesMulti", "orfCategory", "accession_number", "protein_group_num", "entry_name"]) # We don't need these. The dropped columns include a "geneSymbol" column that is a duplicate of the original GeneSymbol.
                df = df.apply(pd.to_numeric) # Now that we've dropped all the extra metadata columns, convert everything to floats.
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name = "Patient_ID"
                self._data["acetylproteomics"] = df

            elif file_name in ["prosp-brca-v3.1-gene-level-cnv-gistic2-all_data_by_genes.gct.gz", 
                                "prosp-brca-v5.4-public-gene-level-cnv-gistic2-all_data_by_genes.gct.gz"]:
                df = pd.read_csv(file_path, sep='\t', skiprows=2, index_col=0, dtype=object) # First two rows of file aren't part of the dataframe. Also, due to extra metadata rows we're going to remove, all cols have mixed types, so we pass dtype=object for now.
                df = df[df["geneSymbol"] != "na"] # There are several metadata rows at the beginning of the dataframe, which duplicate the clinical and derived_molecular dataframes. They all don't have a value for geneSymbol, so we'll use that to filter them out.
                df = df.drop(columns="Cytoband")
                df["geneSymbol"] = df["geneSymbol"].str.rsplit('|', n=1, expand=True)[0] # Some of the geneSymbols have the gene IDs appended to them, to get rid of duplicates. We're going to create a multiindex with all the gene names and gene IDs, so we can drop the appended IDs.
                df = df.rename(columns={"geneSymbol": "Name", "Gene.ID": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])
                df = df.apply(pd.to_numeric) # Now that we've dropped all the extra metadata columns, convert everything to floats.
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name = "Patient_ID"
                self._data["CNV"] = df

            elif file_name in ["prosp-brca-v3.1-phosphoproteome-ratio-norm-NArm.gct.gz", 
                                "prosp-brca-v5.4-public-phosphoproteome-ratio-norm-NArm.gct.gz"]:
                df = pd.read_csv(file_path, sep='\t', skiprows=2, dtype=object) # First two rows of file aren't part of the dataframe. Also, due to extra metadata rows we're going to remove, all cols have mixed types, so we pass dtype=object for now.
                df = df[df["GeneSymbol"] != "na"] # There are several metadata rows at the beginning of the dataframe, which duplicate the clinical and derived_molecular dataframes. They all don't have a value for GeneSymbol, so we'll use that to filter them out.

                # Prepare some columns we'll need later for the multiindex
                df["variableSites"] = df["variableSites"].str.replace(r"[a-z\s]", "", regex=True) # Get rid of all lowercase delimeters and whitespace in the sites
                df = df.rename(columns={
                    "GeneSymbol": "Name",
                    "variableSites": "Site",
                    "sequence": "Peptide", # We take this instead of sequenceVML, to match the other datasets' format
                    "accession_numbers": "Database_ID" # We take all accession numbers they have, instead of the singular accession_number column
                    })

                # Some rows have at least one localized phosphorylation site, but also have other phosphorylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                split_ids = df["id"].str.split('_', expand=True)
                unlocalized_to_drop = df.index[~split_ids[3].eq(split_ids[4]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 3 of the split "id" column is number of phosphorylations detected, and column 4 is number of phosphorylations localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])                

                df = df.drop(columns=["id", "id.description", "geneSymbol", "numColumnsVMsiteObserved", "bestScore", "bestDeltaForwardReverseScore",
                "Best_scoreVML", "Best_numActualVMSites_sty", "Best_numLocalizedVMsites_sty", "sequenceVML",
                "accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA", "protein_mw", "species",
                "speciesMulti", "orfCategory", "accession_number", "protein_group_num", "entry_name"]) # We don't need these. The dropped columns include a "geneSymbol" column that is a duplicate of the original GeneSymbol.
                df = df.apply(pd.to_numeric) # Now that we've dropped all the extra metadata columns, convert everything to floats.
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name = "Patient_ID"
                self._data["phosphoproteomics"] = df

            elif file_name in ["prosp-brca-v3.1-proteome-ratio-norm-NArm.gct.gz", 
                                "prosp-brca-v5.4-public-proteome-ratio-norm-NArm.gct.gz"]:
                df = pd.read_csv(file_path, sep='\t', skiprows=2, dtype=object) # First two rows of file aren't part of the dataframe. Also, due to extra metadata rows we're going to remove, all cols have mixed types, so we pass dtype=object for now.
                df = df[df["GeneSymbol"] != "na"] # There are several metadata rows at the beginning of the dataframe, which duplicate the clinical and derived_molecular dataframes. They all don't have a value for GeneSymbol, so we'll use that to filter them out.

                df = df.rename(columns={"GeneSymbol": "Name", "accession_numbers": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])
                df = df.drop(columns=["id", "id.description", "geneSymbol", "numColumnsProteinObserved", "numSpectraProteinObserved",
                "protein_mw", "percentCoverage", "numPepsUnique", "scoreUnique", "species", "orfCategory", "accession_number", 
                "subgroupNum", "entry_name"]) # We don't need these. The dropped columns include a "geneSymbol" column that is a duplicate of GeneSymbol.
                df = df.apply(pd.to_numeric) # Now that we've dropped all the extra metadata columns, convert everything to floats.
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name = "Patient_ID"
                self._data["proteomics"] = df

            elif file_name in ["prosp-brca-v3.1-rnaseq-fpkm-log2-row-norm-2comp.gct.gz", 
                                "prosp-brca-v5.4-public-rnaseq-fpkm-log2.gct.gz"]:
                df = pd.read_csv(file_path, sep='\t', skiprows=2, index_col=0, dtype=object) # First two rows of file aren't part of the dataframe. Also, due to extra metadata rows we're going to remove, all cols have mixed types, so we pass dtype=object for now.
                df = df[df["geneSymbol"] != "na"] # There are several metadata rows at the beginning of the dataframe, which duplicate the clinical and derived_molecular dataframes. They all don't have a value for GeneSymbol, so we'll use that to filter them out.
                df = df.set_index("geneSymbol")
                df = df.drop(columns="description") # We don't need this.
                df = df.apply(pd.to_numeric) # Now that we've dropped all the extra metadata columns, convert everything to floats.
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name = "Patient_ID"
                self._data["transcriptomics"] = df

            elif file_name == "prosp-brca-v3.1-sample-annotation.csv.gz":
                df = pd.read_csv(file_path, index_col=0)
                df = df.drop(columns="Participant") # This column is just a duplicate of the index
                df = df.rename(columns={"Sample.IDs": "Replicate_Measurement_IDs", "Type": "Sample_Tumor_Normal"})
                df = df.replace("unknown", np.nan)
                df = df.astype({"Age.in.Month": np.float64})
                df.index.name = "Patient_ID"
                
                # Separate the clinical and derived_molecular dataframes
                clinical = df[[
                     "Replicate_Measurement_IDs", "Sample_Tumor_Normal", "Age.in.Month", "Gender", "Race", "Human.Readable.Label", "Experiment", "Channel", "Stage", 
                     "PAM50", "NMF.v2.1", "ER", "PR", "ER.IHC.Score", "PR.IHC.Score", "Coring.or.Excision", "Ischemia.Time.in.Minutes", 
                     "Ischemia.Decade", "Necrosis", "Tumor.Cellularity", "Total.Cellularity", "In.CR", "QC.status"]]
                self._data["clinical"] = clinical

                derived_molecular = df[[
                    "HER2.IHC.Score", "HER2.FISH.Status", "HER2.original", "HER2.Amplified", "HER2.refined", "STARD3.ERBB2.GRB7.protein", 
                    "HER2.class.Satpathy", "HER2.status.Satpathy", "PAM50.Her2.CNA", "PAM50.Her2.HER2.status", "CDH1.mutation", 
                    "GATA3.mutation", "MAP3K1.mutation", "PIK3CA.mutation", "PTEN.mutation", "TP53.mutation", "CDH1.mutation.status", 
                    "GATA3.mutation.status", "MAP3K1.mutation.status", "PIK3CA.mutation.status", "PTEN.mutation.status", "TP53.mutation.status", 
                    "Number.of.Mutations", "Number.of.Mutated.Genes", "Chromosome.INstability.index.CIN.", "ESTIMATE.TumorPurity", 
                    "ESTIMATE.ImmuneScore", "ESTIMATE.StromalScore", "xCell.ImmuneScore", "xCell.StromaScore", "Cibersort.Absolute.score", "Stemness.Score"]]
                self._data["derived_molecular"] = derived_molecular
                
            elif file_name == "prosp-brca-v5.4-public-sample-annotation.csv.gz":
                df = pd.read_csv(file_path, index_col=0)
                df = df.rename(columns={"Sample.IDs": "Replicate_Measurement_IDs", "Tumor.Stage": "Stage"})
                #Note: The information here is the same as before, but the column has a different name. Anyone trying to get this information will have a hard time unless we rename it
                #Or at least make it known in the update notes that there has been a change and let people know about it
                df = df.replace("unknown", np.nan)
                df = df.astype({"Age.in.Month": np.float64})
                df.index.name = "Patient_ID"
                
                #Add tissue type column because it doesn't exist, all tissue should be cancer
                df["Sample_Tumor_Normal"] = "Tumor"
                
                clinical = df[['Replicate_Measurement_IDs', 'Sample_Tumor_Normal', 'TMT.Plex', 'TMT.Channel', 
                   'Stage', 'Ischemia.Time.in.Minutes', 'PAM50', 'NMF.Cluster',
                   'NMF.Cluster.Membership.Score', 'Age.in.Month', 'Gender', 'Ethnicity',
                   'ER.Updated.Clinical.Status', 'PR.Clinical.Status',
                   'ERBB2.Updated.Clinical.Status', 'TNBC.Updated.Clinical.Status',
                   'ERBB2.Proteogenomic.Status', 'TOP2A.Proteogenomic.Status']]
                self._data["clinical"] = clinical
                
                derived_molecular = df[[
                   'ERBB2.Gene.Amplified', 'TOP2A.Gene.Amplified', 'ESTIMATE.TumorPurity', 'CIBERSORT.Absolute.Score', 'ESTIMATE.Immune.Score',
                   'xCell.Immune.Score', 'ESTIMATE.Stromal.Score', 'xCell.Stromal.Score', 'CD3.TILS.Status', 'CD3.TILS.Counts',
                   'Number.of.non.synonymous.Mutations', 'APOBEC.Signature', 'Chromosome.INstability.index.CIN.', 'Stemness.Score', 'TP53.Mutation.Type', 'PIK3CA.Mutation.Type',
                   'PTEN.Mutation.Type', 'MAP3K1.Mutation.Type', 'AKT1.Mutation.Type', 'GATA3.Mutation.Type', 'CBFB.Mutation.Type', 'KMT2C.Mutation.Type', 'SF3B1.Mutation.Type',
                   'ARID1A.Mutation.Type', 'MLLT4.Mutation.Type', 'TP53.Mutation.Status', 'PIK3CA.Mutation.Status', 'PTEN.Mutation.Status',
                   'MAP3K1.Mutation.Status', 'AKT1.Mutation.Status', 'GATA3.Mutation.Status', 'CBFB.Mutation.Status',
                   'KMT2C.Mutation.Status', 'SF3B1.Mutation.Status', 'ARID1A.Mutation.Status', 'MLLT4.Mutation.Status']]
                self._data["derived_molecular"] = derived_molecular

            elif file_name == "Breast_One_Year_Clinical_Data_20160927.xls" and self._version == "3.1.1":
                df = pd.read_excel(file_path)

                # Replace redundant values for "not reported" with NaN
                nan_equivalents = ['Not Reported/ Unknown', 'Reported/ Unknown', 'Not Reported / Unknown', 'Not Reported /Unknown',
                    'Not Applicable', 'not applicable', 'Not applicable;', 'na', 'Not Performed', 'Not Performed;',
                    'Unknown tumor status', 'Unknown Tumor Status','Unknown', 'unknown', 'Not specified', 'Not Reported/ Unknown;']

                df = df.replace(nan_equivalents, np.nan)

                # Set and name the index
                df = df.rename(columns={"Participant ID": "Patient_ID"})
                df["Patient_ID"] = "X" + df["Patient_ID"]
                df = df.set_index("Patient_ID")
                df = df.sort_index()

                self._data["followup"] = df

            elif file_name == "prosp-brca-v3.0-v1.4.somatic.variants.070918.maf.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.rename(columns={"Sample.ID": "Patient_ID"})

                df = df[['Patient_ID','Hugo_Symbol','Variant_Classification','HGVSp_Short']]
                df = df.rename(columns={
                    "Hugo_Symbol":"Gene",
                    "Variant_Classification":"Mutation",
                    "HGVSp_Short":"Location"}) # Rename the columns we want to keep to the appropriate names

                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")

                self._data["somatic_mutation"] = df

            #elif file_name == "prosp-brca-v5.4-public-immune-profiling-scores-combined.gct.gz" and self._version == "5.4":
            #    df = pd.read_csv(file_path, sep='\t', skiprows=2)

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        clinical = self._data["clinical"]
        clinical = clinical.reindex(master_index)

        # Fill in NaNs in the clinical table's Sample_Tumor_Normal column
        clinical["Sample_Tumor_Normal"] = clinical["Sample_Tumor_Normal"].where(cond=~(pd.isnull(clinical["Sample_Tumor_Normal"]) & ~clinical.index.str.endswith(".N")), other="Tumor")

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = clinical

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='breast cancer', pmid=33212010)


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .file_download import download, download_pdc_id, list_pdc_datasets

from .pdcbrca import PdcBrca
from .pdcccrcc import PdcCcrcc
from .pdccoad import PdcCoad
from .pdcgbm import PdcGbm
from .pdchnscc import PdcHnscc
from .pdclscc import PdcLscc
from .pdcluad import PdcLuad
from .pdcov import PdcOv
from .pdcpdac import PdcPdac
from .pdcucec import PdcUcec

from .pancanbrca import PancanBrca
from .pancanccrcc import PancanCcrcc
from .pancancoad import PancanCoad
from .pancangbm import PancanGbm
from .pancanhnscc import PancanHnscc
from .pancanlscc import PancanLscc
from .pancanluad import PancanLuad
from .pancanov import PancanOv
from .pancanucec import PancanUcec
from .pancanpdac import PancanPdac

def list_datasets(print_list=True):
    """Print available datasets in the cptac.pancan module.

    Parameters:
    print_list (bool, optional): Whether to print the list. Default is True. Otherwise, it's returned as a string.
    """

    datasets = [
        "PancanBrca",
        "PancanCcrcc",
        "PancanCoad",
        "PancanGbm",
        "PancanHnscc",
        "PancanLscc",
        "PancanLuad",
        "PancanOv",
        "PancanUcec",
        "PancanPdac"
    ]

    str_result = "\n".join(datasets)

    if print_list:
        print(str_result)
    else:
        return str_result



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime

from cptac.dataset import Dataset
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError

class PdcLscc(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": [
                "acetylome.tsv.gz",
                "clinical.tsv.gz",
                "phosphoproteome.tsv.gz",
                "proteome.tsv.gz",
                "ubiquitylome.tsv.gz",
                "aliquot_to_patient_ID.tsv"
            ]
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="pdclscc", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet, attempt_update_index=False)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below

            if file_name == "clinical.tsv.gz":
                df = pd.read_csv(file_path, sep="\t", index_col=0)
                clin_drop_rows = ['JHU HNSCC CR', 'LSCC Global CR', 'LSCC KGG CR', 'LSCC Tumor ONLY CR', 
                             'LUAD Global CR (pool 1)', 'LUAD Global CR (pool 2)', 'LUAD KGG CR']
                df = df.drop(clin_drop_rows, axis = 'index') 
                self._data["clinical"] = df

            if file_name == "acetylome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                self._data["acetylproteomics"] = df

            if file_name == "phosphoproteome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                self._data["phosphoproteomics"] = df

            if file_name == "proteome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                self._data["proteomics"] = df

            if file_name == "ubiquitylome.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                self._data["ubiquitylomics"] = df
                                
            # aliquot_to_patient_ID.tsv contains only unique aliquots (no duplicates), 
            # so there is no need to slice out cancer specific aliquots
            elif file_name == "aliquot_to_patient_ID.tsv":
                df = pd.read_csv(file_path, sep = "\t", index_col = 'aliquot_ID', usecols = ['aliquot_ID', 'patient_ID'])
                map_dict = df.to_dict()['patient_ID'] # create dictionary with aliquot_ID as keys and patient_ID as values
                self._helper_tables["map_ids"] = map_dict

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = f"Formatting {self.get_cancer_type()} dataframes..."
        print(formatting_msg, end='\r')

        # Common rows to drop 
        drop_rows = ['JHU HNSCC CR', 'LSCC Tumor ONLY CR', 'LUAD Global CR (pool 1)', 'LUAD Global CR (pool 1)',
                     'LUAD Global CR (pool 2)']        
        
        # Get dictionary with aliquot_ID as keys and patient_ID as values
        mapping_dict = self._helper_tables["map_ids"]
        
        # Proteomics
        prot = self._data["proteomics"]
        prot['Patient_ID'] = prot['aliquot_submitter_id'].replace(mapping_dict) # aliquots to patient IDs (normals have '.N')
        prot = prot.set_index('Patient_ID')
        prot = prot.drop(['aliquot_submitter_id', 'case_submitter_id'], axis = 'columns')
        self._data["proteomics"] = prot
        
        # Phosphoproteomics
        phos = self._data["phosphoproteomics"]
        phos['Patient_ID'] = phos['aliquot_submitter_id'].replace(mapping_dict) # aliquots to patient IDs (normals have '.N')
        phos = phos.set_index('Patient_ID')
        phos = phos.drop(['aliquot_submitter_id', 'case_submitter_id'], axis = 'columns')        
        phos = phos.drop(drop_rows, axis = 'index')
        phos = map_database_to_gene_pdc(phos, 'refseq') # map refseq IDs to gene names
        self._data["phosphoproteomics"] = phos
        
        # Acetylproteomics
        acetyl = self._data["acetylproteomics"]
        acetyl['Patient_ID'] = acetyl['aliquot_submitter_id'].replace(mapping_dict) # aliquots to patient IDs (normals have '.N')
        acetyl = acetyl.set_index('Patient_ID')
        acetyl = acetyl.drop(['aliquot_submitter_id', 'case_submitter_id'], axis = 'columns') 
        acetyl = acetyl.drop(drop_rows, axis = 'index')
        acetyl = map_database_to_gene_pdc(acetyl, 'refseq') # map refseq IDs to gene names
        self._data["acetylproteomics"] = acetyl
        
        # Ubiquitylomics
        ubiq = self._data["ubiquitylomics"]
        ubiq['Patient_ID'] = ubiq['aliquot_submitter_id'].replace(mapping_dict) # aliquots to patient IDs (normals have '.N')
        ubiq = ubiq.set_index('Patient_ID')
        ubiq = ubiq.drop(['aliquot_submitter_id', 'case_submitter_id'], axis = 'columns')
        ubiq = ubiq.drop(['LUAD KGG CR'], axis = 'index') 
        ubiq = map_database_to_gene_pdc(ubiq, 'refseq') # map refseq IDs to gene names
        self._data["ubiquitylomics"] = ubiq
        
        
        # Sort rows (tumor first then normal) and columns by first level (protein/gene name)
        self._data = sort_all_rows_pancan(self._data)


        print(" " * len(formatting_msg), end='\r') # Erase the formatting message



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import pandas as pd
import requests
import shutil
import warnings

import cptac
from cptac.file_download import get_box_token
from cptac.exceptions import DatasetAlreadyInstalledWarning, InvalidParameterError, NoInternetError, PdcDownloadError

from .pancanbrca import SOURCES as BRCA_SOURCES
from .pancanccrcc import SOURCES as CCRCC_SOURCES
from .pancancoad import SOURCES as COAD_SOURCES
from .pancangbm import SOURCES as GBM_SOURCES
from .pancanhnscc import SOURCES as HNSCC_SOURCES
from .pancanlscc import SOURCES as LSCC_SOURCES
from .pancanluad import SOURCES as LUAD_SOURCES
from .pancanov import SOURCES as OV_SOURCES
from .pancanucec import SOURCES as UCEC_SOURCES
from .pancanpdac import SOURCES as PDAC_SOURCES

STUDY_IDS_MAP = {
    "pdcbrca": {
        "acetylome": "PDC000239", # Prospective Breast BI Acetylome
        "phosphoproteome": "PDC000121", # Prospective BRCA Phosphoproteome S039-2
        "proteome": "PDC000120", # Prospective BRCA Proteome S039-1
    },
    "pdcccrcc": {
        "phosphoproteome": "PDC000128", # CPTAC CCRCC Discovery Study - Phosphoproteme S044-2
        "proteome": "PDC000127", # CPTAC CCRCC Discovery Study - Proteome S044-1
    },
    "pdccoad": {
        "phosphoproteome": "PDC000117", # Prospective COAD Phosphoproteome S037-3
        "proteome": "PDC000116", # Prospective COAD Proteome S037-2
    },
    "pdcgbm": {
        "acetylome": "PDC000245", # CPTAC GBM Discovery Study - Acetylome
        "phosphoproteome": "PDC000205", # CPTAC GBM Discovery Study - Phosphoproteome
        "proteome": "PDC000204", # CPTAC GBM Discovery Study - Proteome
    },
    "pdchnscc": {
        "phosphoproteome": "PDC000222", # CPTAC HNSCC Discovery Study - Phosphoproteome
        "proteome": "PDC000221", # CPTAC HNSCC Discovery Study - Proteome
    },
    "pdclscc": {
        "acetylome": "PDC000233", # CPTAC LSCC Discovery Study - Acetylome
        "phosphoproteome": "PDC000232", # CPTAC LSCC Discovery Study - Phosphoproteome
        "proteome": "PDC000234", # CPTAC LSCC Discovery Study - Proteome
        "ubiquitylome": "PDC000237", # CPTAC LSCC Discovery Study - Ubiquitylome
    },
    "pdcluad": {
        "acetylome": "PDC000224", # CPTAC LUAD Discovery Study - Acetylome
        "phosphoproteome": "PDC000149", # CPTAC LUAD Discovery Study - Phosphoproteome
        "proteome": "PDC000153", # CPTAC LUAD Discovery Study - Proteome
    },
    "pdcov": {
        "phosphoproteome": "PDC000119", # Prospective OV Phosphoproteome S038-3
        "proteome": "PDC000118", # Prospective OV Proteome S038-2
    },
    "pdcpdac": {
        "proteome": "PDC000270", # CPTAC PDAC Discovery Study - Proteome
        "phosphoproteome": "PDC000271", # CPTAC PDAC Discovery Study - Phosphoproteome
    },
    "pdcucec": {
        "acetylome": "PDC000226", # CPTAC UCEC Discovery Study - Acetylome
        "phosphoproteome": "PDC000126", # UCEC Discovery - Phosphoproteome S043-2
        "proteome": "PDC000125", # UCEC Discovery - Proteome S043-1
    },
}


def download(dataset, version="latest", redownload=False):

    dataset = dataset.lower()

    if dataset.startswith("pdc"):
        box_token = get_box_token()
        if dataset != 'pdcbrca': # pdcbrca is the only dataset that doesn't need a mapping file for PDC
            mapping = cptac.download(dataset, version=version, redownload=redownload, _box_auth=True, _box_token=box_token) # download helper file for mapping aliquots to patient IDs        
            omics = _pdc_download(dataset, version=version, redownload=redownload)        
            if omics and mapping:
                return True
            else:
                return False
        else: # pdcbrca only needs omics
            omics = _pdc_download(dataset, version=version, redownload=redownload)
            if omics:
                return True
            else:
                return False

    elif dataset.startswith("pancan") or dataset == "all":
        box_token = get_box_token()

        if dataset == "pancanbrca":
            sources = BRCA_SOURCES
        elif dataset == "pancanccrcc":
            sources = CCRCC_SOURCES
        elif dataset == "pancancoad":
            sources = COAD_SOURCES
        elif dataset == "pancangbm":
            sources = GBM_SOURCES
        elif dataset == "pancanhnscc":
            sources = HNSCC_SOURCES
        elif dataset == "pancanlscc":
            sources = LSCC_SOURCES
        elif dataset == "pancanluad":
            sources = LUAD_SOURCES
        elif dataset == "pancanov":
            sources = OV_SOURCES
        elif dataset == "pancanucec":
            sources = UCEC_SOURCES
        elif dataset == "pancanpdac":
            sources = PDAC_SOURCES
        elif dataset == "all":
            sources = sorted(set(BRCA_SOURCES + CCRCC_SOURCES + COAD_SOURCES + GBM_SOURCES + HNSCC_SOURCES + LSCC_SOURCES + LUAD_SOURCES + OV_SOURCES + UCEC_SOURCES + PDAC_SOURCES))
        else:
            raise InvalidParameterError(f"{dataset} is not a valid dataset.")

        overall_success = True
        for source in sources:

            if source.startswith("pdc"):
                single_success = download(source, version=version, redownload=redownload)
            else:
                single_success = cptac.download(source, version=version, redownload=redownload, _box_auth=True, _box_token=box_token)

            if not single_success:
                overall_success = False

        return overall_success

    else:
        return cptac.download(dataset, version=version, redownload=redownload, _box_auth=True)

def download_pdc_id(pdc_id, _download_msg=True):
    """Download a PDC dataset by its PDC study id.
    
    Returns:
    pandas.DataFrame: The clinical table for the study id.
    pandas.DataFrame: The quantitative table for the study id.
    """

    if _download_msg:
        clin_msg = f"Downloading clinical table for {pdc_id}..."
        print(clin_msg, end="\r")

    # Download the clinical table
    clin = _download_study_clin(pdc_id).\
    set_index("case_submitter_id").\
    sort_index()

    if _download_msg:
        print(" " * len(clin_msg), end="\r")
        bio_msg = f"Downloading biospecimenPerStudy table for {pdc_id}..."
        print(bio_msg, end="\r")

    # The the biospecimenPerStudy table, which has both patient IDs and aliquot IDs
    bio = _download_study_biospecimen(pdc_id).\
    set_index("aliquot_submitter_id").\
    sort_index()

    if _download_msg:
        print(" " * len(bio_msg), end="\r")
        quant_msg = f"Downloading quantitative table for {pdc_id}..."
        print(quant_msg, end="\r")

    # Get the quantitative data table
    quant = _download_study_quant(pdc_id)

    if _download_msg:
        print(" " * len(quant_msg), end="\r")
        format_msg = f"Formatting tables for {pdc_id}..."
        print(format_msg, end="\r")

    # Join the patient IDs from the biospecimenPerStudy table into the quant table
    quant = quant.\
    assign(aliquot_submitter_id=quant.iloc[:, 0].str.split(":", n=1, expand=True)[1]).\
    drop(columns=quant.columns[0]).\
    set_index("aliquot_submitter_id").\
    sort_index()

    quant = bio.\
    join(quant, how="inner").\
    reset_index().\
    set_index(["case_submitter_id", "aliquot_submitter_id"]).\
    sort_index()

    # Clear message
    if _download_msg:
        print(" " * len(format_msg), end="\r")

    return clin, quant

def list_pdc_datasets():
    for dataset in STUDY_IDS_MAP.keys():
        print(f"Pdc{dataset[3:].title()}:")
        for data_type in STUDY_IDS_MAP[dataset].keys():
            print(f"\t{data_type}: {STUDY_IDS_MAP[dataset][data_type]}")

# Helper functions
    
def _pdc_download(dataset, version, redownload):
    """Download data for the specified cancer type from the PDC."""

    dataset = str.lower(dataset)

    if dataset == "pdcall":
        overall_result = True
        for dataset in STUDY_IDS_MAP.keys():
            if not pdc_download(dataset, version, redownload):
                overall_result = False

        return overall_result

    if not dataset.startswith("pdc"):
        raise InvalidParameterError(f"pdc_download function can only be used for PDC datasets, which start with the prefix 'pdc'. You tried to download '{dataset}'.")

    if dataset not in STUDY_IDS_MAP.keys():
        raise InvalidParameterError(f"PDC dataset must be one of the following:\n{list(STUDY_IDS_MAP.keys())}\nYou passed '{dataset}'.")

    dataset_ids = STUDY_IDS_MAP[dataset]

    # Get the directory to where to store the data, and see if it exists
    path_here = os.path.abspath(os.path.dirname(__file__))
    cancer_dir = os.path.join(path_here, f"data_{dataset}")

    if os.path.isdir(cancer_dir):

        index_path = os.path.join(cancer_dir, "index.txt")

        # Check that they also have the index
        if not os.path.isfile(index_path):
            redownload = True
        else:
            # The PDC doesn't have a versioning scheme for the tables they serve, so originally we just called it version 0.0 but later decided it would be better to call it 1.0. So, check if theirs is called 0.0; if so, replace it with 1.0.

            with open(index_path, "r") as index_file:
                first_line = index_file.readline()

            if first_line.startswith("#0.0"):
                redownload=True

        if redownload:
            shutil.rmtree(cancer_dir)
        else:

            return True

    os.mkdir(cancer_dir)
    data_dir = os.path.join(cancer_dir, f"{dataset}_v1.0")
    os.mkdir(data_dir)

    # We'll combine all the clinical tables in case there are differences
    master_clin = pd.DataFrame()

    for data_type in dataset_ids.keys():

        # Print an update
        download_msg = f"Downloading {dataset} {data_type} files..."
        print(download_msg, end="\r")

        # Get the clinical and quantitative tables for the study ID
        clin, quant = download_pdc_id(dataset_ids[data_type], _download_msg=False)

        # Print a new update
        print(" " * len(download_msg), end="\r")
        save_msg = f"Saving {dataset} {data_type} files..."
        print(save_msg, end="\r")

        # Append the clinical dataframe
        master_clin = master_clin.append(clin)

        # Save the quantitative table
        quant.to_csv(os.path.join(data_dir, f"{data_type}.tsv.gz"), sep="\t")

        # Erase update
        print(" " * len(save_msg), end="\r")

    # Print an update
    save_msg = f"Saving {dataset} clinical file..."
    print(save_msg, end="\r")

    # Drop any duplicated rows in combined clinical table, then save it too
    master_clin = master_clin.drop_duplicates(keep="first")

    master_clin.to_csv(os.path.join(data_dir, "clinical.tsv.gz"), sep="\t")

    # Write a dummy index with just version numbers
    index_path = os.path.join(cancer_dir, "index.txt")

    with open(index_path, "w") as index_file:
        index_file.write("#1.0\n")
        
    # Erase update
    print(" " * len(save_msg), end="\r")

    return True

def _download_study_clin(pdc_study_id):
    """Download PDC clinical data for a particular study."""

    clinical_query = '''
    query {
        clinicalPerStudy(pdc_study_id: "''' + pdc_study_id + '''", acceptDUA: true) {
            age_at_diagnosis, ajcc_clinical_m, ajcc_clinical_n, ajcc_clinical_stage, ajcc_clinical_t, ajcc_pathologic_m,
            ajcc_pathologic_n, ajcc_pathologic_stage, ajcc_pathologic_t, ann_arbor_b_symptoms, ann_arbor_clinical_stage,
            ann_arbor_extranodal_involvement, ann_arbor_pathologic_stage, best_overall_response, burkitt_lymphoma_clinical_variant,
            case_id, case_submitter_id, cause_of_death, circumferential_resection_margin, classification_of_tumor, colon_polyps_history,
            days_to_best_overall_response, days_to_birth, days_to_death, days_to_diagnosis, days_to_hiv_diagnosis, days_to_last_follow_up,
            days_to_last_known_disease_status, days_to_new_event, days_to_recurrence, demographic_id, demographic_submitter_id,
            diagnosis_id, diagnosis_submitter_id, disease_type, ethnicity, figo_stage, gender, hiv_positive, hpv_positive_type, hpv_status,
            icd_10_code, iss_stage, last_known_disease_status, laterality, ldh_level_at_diagnosis, ldh_normal_range_upper,
            lymphatic_invasion_present, lymph_nodes_positive, method_of_diagnosis, morphology, new_event_anatomic_site, new_event_type,
            overall_survival, perineural_invasion_present, primary_diagnosis, primary_site, prior_malignancy, prior_treatment,
            progression_free_survival, progression_free_survival_event, progression_or_recurrence, race, residual_disease,
            site_of_resection_or_biopsy, status, synchronous_malignancy, tissue_or_organ_of_origin, tumor_cell_content, tumor_grade,
            tumor_stage, vascular_invasion_present, vital_status, year_of_birth, year_of_death, year_of_diagnosis
        }
    }
    '''

    result_json = _query_pdc(clinical_query)
    result_df = pd.\
    DataFrame(result_json["data"]["clinicalPerStudy"])

    return result_df

def _download_study_biospecimen(pdc_study_id):
    """Download PDC biospecimen data for a particular study."""

    biospecimen_query = '''
    query {
        biospecimenPerStudy(pdc_study_id: "''' + pdc_study_id + '''", acceptDUA: true) {
            aliquot_submitter_id
            case_submitter_id
        }
    }
    '''

    result_json = _query_pdc(biospecimen_query)
    result_df = pd.\
    DataFrame(result_json["data"]["biospecimenPerStudy"])

    return result_df

def _download_study_quant(pdc_study_id):
    """Download PDC quantitative data for a particular study."""

    proteome_query = '''
    query {
        quantDataMatrix(pdc_study_id: "''' + pdc_study_id + '''", data_type: "log2_ratio", acceptDUA: true)
    }
    '''

    result_json = _query_pdc(proteome_query)
    result_df = pd.DataFrame(result_json["data"]["quantDataMatrix"])

    if result_df.shape[1] != 0:
        result_df = result_df.set_index(result_df.columns[0]).transpose()
    else:
        raise PdcDownloadError(f"quantDataMatrix table returned for PDC study ID {pdc_study_id} was empty.")

    return result_df

def _query_pdc(query):
    """Send a GraphQL query to the PDC and return the results."""

    url = 'https://pdc.cancer.gov/graphql'

    try:
        response = requests.post(url, json={'query': query})
        response.raise_for_status() # Raises a requests.HTTPError if the response code was unsuccessful

    except requests.RequestException: # Parent class for all exceptions in the requests module
        raise NoInternetError("Insufficient internet. Check your internet connection.") from None

    return response.json()

def _check_ids_match(ids_map):
    """Check that the ids in the download function's STUDY_IDS_MAP match up."""
    
    for cancer in ids_map.values():
        for data in cancer.values():
            pdc_study_id = data["pdc_study_id"]
            study_submitter_id = data["study_submitter_id"]

            query = '''
            query {
              study (pdc_study_id: "''' + pdc_study_id + '''" acceptDUA: true) {
                pdc_study_id,
                study_submitter_id
              }
            }
            '''

            idres = _query_pdc(query)

            server_psi = idres["data"]["study"][0]["pdc_study_id"]
            server_ssi = idres["data"]["study"][0]["study_submitter_id"]

            assert server_psi == pdc_study_id
            assert server_ssi == study_submitter_id

            print(f"{server_psi} == {pdc_study_id}")
            print(f"{server_ssi} == {study_submitter_id}")
            print()



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime

from cptac.dataset import Dataset
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError
from cptac.utils import get_boxnote_text


class UmichCcrcc(Dataset):

    def __init__(self, no_internet, version):
        """Load all of the umichccrcc dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": ["Report_abundance_groupby=protein_protNorm=MD_gu=2.tsv",
                    "Report_abundance_groupby=multi-site_protNorm=MD_gu=2.tsv",
                    "aliquot_to_patient_ID.tsv",
                    "README_v3.boxnote", # proteomics 
                    "README.boxnote" # phosphoproteomics
                #"S039_BCprospective_observed_0920.tsv.gz",
                #"S039_BCprospective_imputed_0920.tsv.gz"
            ]
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="umichccrcc", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below
            
            
            if file_name == "Report_abundance_groupby=protein_protNorm=MD_gu=2.tsv":
                df = pd.read_csv(file_path, sep = "\t") 
                df['Database_ID'] = df.Index.apply(lambda x: x.split('|')[0]) # get protein identifier 
                df['Name'] = df.Index.apply(lambda x: x.split('|')[6]) # get protein name 
                df = df.set_index(['Name', 'Database_ID']) # set multiindex
                df = df.drop(columns = ['Index', 'MaxPepProb', 'NumberPSM', 'Gene']) # drop unnecessary  columns
                df = df.transpose()
                ref_intensities = df.loc["ReferenceIntensity"] # get reference intensities to use to calculate ratios 
                df = df.subtract(ref_intensities, axis="columns") # subtract reference intensities from all the values
                df = df.iloc[1:,:] # drop ReferenceIntensity row 
                df.index.name = 'Patient_ID'
                self._data["proteomics"] = df
 
                
            elif file_name == "Report_abundance_groupby=multi-site_protNorm=MD_gu=2.tsv":
                df = pd.read_csv(file_path, sep = "\t") 
                # Parse a few columns out of the "Index" column that we'll need for our multiindex
                df[['Database_ID','Transcript_ID',"Gene_ID","Havana_gene","Havana_transcript","Transcript","Name","Site"]] = df.Index.str.split("\\|",expand=True)
                df[['num1','start',"end","detected_phos","localized_phos","Site"]] = df.Site.str.split("_",expand=True) 

                # Some rows have at least one localized phosphorylation site, but also have other
                # phosphorylations that aren't localized. We'll drop those rows, if their localized sites 
                # are duplicated in another row, to avoid creating duplicates, because we only preserve 
                # information about the localized sites in a given row. However, if the localized sites aren't 
                # duplicated in another row, we'll keep the row.
                unlocalized_to_drop = df.index[~df["detected_phos"].eq(df["localized_phos"]) & \
                                               df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)]
                # dectected_phos of the split "Index" column is number of phosphorylations detected, and 
                # localized_phos is number of phosphorylations localized, so if the two values aren't equal, 
                # the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                df = df[df['Site'].notna()] # only keep columns with phospho site 
                df = df.set_index(['Name', 'Site', 'Peptide', 'Database_ID']) # create a multiindex in this order
                #drop columns not needed in df 
                df.drop(['Gene', "Index", "num1", "start", "end", "detected_phos", "localized_phos", "Havana_gene", 
                         "Havana_transcript", "MaxPepProb", "Gene_ID","Transcript_ID", "Transcript"], axis=1, inplace=True)
                df = df.transpose()
                ref_intensities = df.loc["ReferenceIntensity"]# Get reference intensities (prep to calculate ratios) 
                df = df.subtract(ref_intensities, axis="columns") # Subtract refintensities from all the values, to get ratios
                df = df.iloc[1:,:] # drop ReferenceIntensity row 
                self._data["phosphoproteomics"] = df
                
                
            # aliquot_to_patient_ID.tsv contains only unique aliquots (no duplicates), 
            # so there is no need to slice out cancer specific aliquots
            # This file can be found on Box under CPTAC/cptac/pancan/helper_files
            elif file_name == "aliquot_to_patient_ID.tsv":
                df = pd.read_csv(file_path, sep = "\t", index_col = 'aliquot_ID', usecols = ['aliquot_ID', 'patient_ID'])
                map_dict = df.to_dict()['patient_ID'] # create dictionary with aliquot_ID as keys and patient_ID as values
                self._helper_tables["map_ids"] = map_dict
                
            elif file_name == "README_v3.boxnote":
                self._readme_files["readme_proteomics"] = get_boxnote_text(file_path)
                
            elif file_name == "README.boxnote":
                self._readme_files["readme_phosphoproteomics"] = get_boxnote_text(file_path)
      
            '''
            elif file_name == "S039_BCprospective_observed_0920.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                df = df.transpose()
                df.index.name = 'Patient_ID'
                df.columns.name = 'Name'
                df = average_replicates(df)
                df = df.sort_values(by=["Patient_ID"])
                self._data["proteomics"] = df  
                
            elif file_name == "S039_BCprospective_imputed_0920.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")
                df = df.transpose()
                df.index.name = 'Patient_ID'
                df.columns.name = 'Name'
                df = average_replicates(df)
                df = df.sort_values(by=["Patient_ID"])
                self._data["proteomics_imputed"] = df'''
            

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')
        
        drop_cols = ['NCI7-1','NCI7-2','NCI7-3','NCI7-4','NCI7-5', 'QC1', 'QC2', 'QC3', 'QC4', 'QC5', 'QC6', 'QC7', 
                    'QC8', 'RefInt_pool01', 'RefInt_pool02', 'RefInt_pool03', 'RefInt_pool04', 'RefInt_pool05', 
                    'RefInt_pool06', 'RefInt_pool07', 'RefInt_pool08', 'RefInt_pool09', 'RefInt_pool10', 'RefInt_pool11', 
                    'RefInt_pool12', 'RefInt_pool13', 'RefInt_pool14', 'RefInt_pool15', 'RefInt_pool16', 'RefInt_pool17', 
                    'RefInt_pool18', 'RefInt_pool19', 'RefInt_pool20', 'RefInt_pool21', 'RefInt_pool22', 'RefInt_pool23']
        
        
        # Get dictionary with aliquot_ID as keys and patient_ID as values
        mapping_dict = self._helper_tables["map_ids"]
        
        # Proteomics
        prot = self._data["proteomics"]
        prot = prot.drop(drop_cols, axis = 'index') # drop quality control and ref intensity cols
        prot = prot.rename(index = mapping_dict) # replace aliquot_IDs with Patient_IDs (normal samples have .N appended)
        self._data["proteomics"] = prot
        
        # Phosphoproteomics 
        phos = self._data["phosphoproteomics"] 
        phos = phos.drop(drop_cols, axis = 'index') # drop quality control and ref intensity cols
        phos = phos.rename(index = mapping_dict) # replace aliquot_IDs with Patient_IDs (normal samples have .N appended)  
        self._data["phosphoproteomics"] = phos
        
        
        # Sort rows (tumor first then normal) and columns by first level (protein/gene name)
        self._data = sort_all_rows_pancan(self._data) 


        print(" " * len(formatting_msg), end='\r') # Erase the formatting message



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, ReindexMapError

class Colon(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the colon dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["0.0", "0.0.1"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "0.0": [
                "clinical.tsi.gz",
                "miRNA.cct.gz",
                "mutation_binary.cbt.gz",
                "mutation.txt.gz",
                "phosphoproteomics_normal.gz",
                "phosphoproteomics_tumor.gz",
                "proteomics_normal.cct.gz",
                "proteomics_tumor.cct.gz",
                "transcriptomics.gz"],
            "0.0.1": [
                "clinical.tsi.gz",
                "Colon_One_Year_Clinical_Data_20160927.xls",
                "Human__CPTAC_COAD__VU__SCNA__ExomeSeq__01_28_2016__BCM__Gene__BCM_CopyWriteR_GISTIC2.cct.gz",
                "miRNA.cct.gz",
                "mutation_binary.cbt.gz",
                "mutation.txt.gz",
                "phosphoproteomics_normal.gz",
                "phosphoproteomics_tumor.gz",
                "proteomics_normal.cct.gz",
                "proteomics_tumor.cct.gz",
                "transcriptomics.gz"],
        }

        super().__init__(cancer_type="colon", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file
            file_name_split = file_name.split(".")
            df_name = file_name_split[0] # Our dataframe name will be the first section of file name (i.e. proteomics.txt.gz becomes proteomics)

            if file_name == 'Colon_One_Year_Clinical_Data_20160927.xls' and self._version == "0.0.1":
                df = pd.read_excel(file_path)

                # Replace redundant values for "not reported" with NaN
                nan_equivalents = ['Not Reported/ Unknown', 'Reported/ Unknown', 'Not Applicable',
                    'na', 'unknown', 'Not Performed', 'Unknown tumor status']

                df = df.replace(nan_equivalents, np.nan)

                # Rename and set index
                df = df.rename(columns={'PPID': 'Patient_ID'})
                df = df.set_index("Patient_ID")
                df = df.sort_index()

                self._data["followup"] = df

            elif file_name == "Human__CPTAC_COAD__VU__SCNA__ExomeSeq__01_28_2016__BCM__Gene__BCM_CopyWriteR_GISTIC2.cct.gz" and self._version == "0.0.1":
                df = pd.read_csv(file_path, sep="\t",index_col=0)
                df = df.sort_index()
                df = df.transpose()
                self._data["CNV"] = df

            else:
                df = pd.read_csv(file_path, sep="\t",index_col=0)
                df = df.sort_index()
                df = df.transpose()
                self._data[df_name] = df # Maps dataframe name to dataframe. self._data was initialized when we called the parent class __init__()

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Reformat and rename the somatic_mutation dataframe
        mut = self._data["mutation"]
        mut = mut.transpose() # Transpose it back to its original orientation
        mut = mut.sort_values(by="SampleID")
        mut = mut.reset_index()
        mut = mut[["SampleID","Gene","Variant_Type","Protein_Change"]]
        mut = mut.drop_duplicates(keep="first") # Get rid of rows that are now duplicates since we didn't keep the mRNA column. We do this before setting the index, because drop_duplicates doesn't consider the index.
        mut = mut.rename(columns={"SampleID":"Patient_ID", "Variant_Type":"Mutation", "Protein_Change":"Location"})
        mut = mut.sort_values(by=["Patient_ID", "Gene"])
        mut = mut.set_index("Patient_ID") # We only do this after the drop_duplicates call above because drop_duplicates doesn't consider the index, but we of course want the Patient_ID to be considered when identifying duplicate rows to drop.
        self._data["somatic_mutation"] = mut # Maps dataframe name to dataframe. self._data was initialized when we called the parent class __init__()
        del self._data["mutation"] # Delete the old version with the old name

        # Rename mutation_binary dataframe to somatic_mutation_binary
        self._data["somatic_mutation_binary"] = self._data["mutation_binary"]
        del self._data["mutation_binary"]

        # Separate clinical and derived molecular dataframes
        all_clinical_data = self._data.get("clinical")
        clinical_df = all_clinical_data.drop(columns=['StromalScore', 'ImmuneScore', 'ESTIMATEScore', 'TumorPurity','immuneSubtype', 'CIN', 'Integrated.Phenotype', 'Transcriptomic_subtype', 'Proteomic_subtype', 'mutation_rate', 'Mutation_Phenotype'])
        derived_molecular_df = all_clinical_data[['StromalScore', 'ImmuneScore', 'ESTIMATEScore', 'TumorPurity', 'immuneSubtype', 'CIN', 'Integrated.Phenotype', 'Transcriptomic_subtype', 'Proteomic_subtype', 'mutation_rate', 'Mutation_Phenotype']]

        # Format the dataframes
        clinical_df = clinical_df.apply(pd.to_numeric, errors="ignore")
        derived_molecular_df = derived_molecular_df.apply(pd.to_numeric, errors="ignore")
        derived_molecular_df = derived_molecular_df.sort_index(axis="columns")

        # Put them in our data dictionary
        self._data["clinical"] = clinical_df # Replaces original clinical dataframe
        self._data["derived_molecular"] = derived_molecular_df

        # Combine the two proteomics dataframes
        prot_tumor = self._data.get("proteomics_tumor")
        prot_normal = self._data.get("proteomics_normal") # Normal entries are already marked with 'N' on the end of the ID
        prot_combined = prot_tumor.append(prot_normal)
        self._data["proteomics"] = prot_combined
        del self._data["proteomics_tumor"]
        del self._data["proteomics_normal"]

        # Get phosphoproteomics dataframes, so we can process and combine them
        phos_tumor = self._data.get("phosphoproteomics_tumor")
        phos_normal = self._data.get("phosphoproteomics_normal") # Normal entries are not marked

        # Mark entries in phosphoproteomics_normal dataframe with an N at the end of the ID
        phos_normal = phos_normal.set_index(phos_normal.index + 'N')

        # Combine the two phosphoproteomics dataframes into one dataframe
        phos_combined = phos_tumor.append(phos_normal)

        # Create our phosphoproteomics columns multiindex
        multiindex = phos_combined.columns.str.split('[_:]', expand=True) # Split the column names into their constituent parts
        multiindex = multiindex.droplevel([2, 4]) # The third level is just empty strings, and the fifth is a duplicate of the second
        multiindex = multiindex.set_names(["Name", "Site", "Database_ID"])
        phos_combined.columns = multiindex
        phos_combined = phos_combined.sort_index(axis=1) # Put all the columns in alphabetical order
        self._data['phosphoproteomics'] = phos_combined
        del self._data["phosphoproteomics_tumor"]
        del self._data["phosphoproteomics_normal"]

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        master_clinical = self._data['clinical'].reindex(master_index)

        # Add a column called Sample_Tumor_Normal to the clinical dataframe indicating whether each sample is a tumor or normal sample. Samples with a Patient_ID ending in N are normal.
        clinical_status_col = generate_sample_status_col(master_clinical, normal_test=lambda sample: sample[-1] == 'N')
        master_clinical.insert(0, "Sample_Tumor_Normal", clinical_status_col)

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = master_clinical 

        # Edit the format of the Patient_IDs to have normal samples marked the same way as in other datasets
        # Currently, normal patient IDs have an "N" appended. We're going to make that a ".N"
        self._data = reformat_normal_patient_ids(self._data, existing_identifier="N", existing_identifier_location="end")

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    # Overload the default how_to_cite function, to provide the specific publication information for the Colon dataset
    def how_to_cite(self):
        """Print instructions for citing the data."""
        super().how_to_cite(cancer_type='colorectal cancer', pmid=31031003)